#!/usr/bin/python3
# -*- coding: utf-8 -*-

#############################################################################
##                                                                         ##
## This file is part of DPAPIck                                            ##
## Windows DPAPI decryption & forensic toolkit                             ##
##                                                                         ##
##                                                                         ##
## Copyright (C) 2010, 2011 Cassidian SAS. All rights reserved.            ##
## Copyright (C) 2023       Insecurity. All rights reserved.               ##
##                                                                         ##
##  Author:  Jean-Michel Picod <jmichel.p@gmail.com>                       ##
##  Updated: Photubias <info@insecurity.be>                                ##
##                                                                         ##
## This program is distributed under GPLv3 licence (see LICENCE.txt)       ##
##                                                                         ##
#############################################################################

import hashlib, os, re, binascii, pickle
from collections import defaultdict
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_v1_5
import dpapick3.crypto as crypto
import dpapick3.eater as eater
import dpapick3.credhist as credhist

class MasterKey(eater.DataStruct):
    """This class represents a MasterKey block contained in a MasterKeyFile"""

    def __init__(self, raw=None):
        self.decrypted = False
        self.key = None
        self.key_hash = None
        self.hmacSalt = None
        self.hmac = None
        self.hmacComputed = None
        self.cipherAlgo = None
        self.hashAlgo = None
        self.rounds = None
        self.iv = None
        self.test = 123
        self.version = None
        self.ciphertext = None
        eater.DataStruct.__init__(self, raw)

    def __getstate__(self):
        d = dict(self.__dict__)
        for k in ["cipherAlgo", "hashAlgo"]:
            if k in d:
                d[k] = d[k].algnum
        return d

    def __setstate__(self, d):
        for k in ["cipherAlgo", "hashAlgo"]:
            if k in d:
                d[k] = crypto.CryptoAlgo(d[k])
        self.__dict__.update(d)

    def parse(self, data):
        self.version = data.eat("L")
        self.iv = data.eat("16s")
        self.rounds = data.eat("L")
        self.hashAlgo = crypto.CryptoAlgo(data.eat("L"))
        self.cipherAlgo = crypto.CryptoAlgo(data.eat("L"))
        self.ciphertext = data.remain()

    def decryptWithHash(self, userSID, pwdhash):
        """Decrypts the masterkey with the given user's hash and SID.
        Simply computes the corresponding key then calls self.decryptWithKey()

        """
        self.decryptWithKey(crypto.derivePwdHash(pwdhash, userSID))

    def decryptWithHash10(self, userSID, pwdhash):
        """Decrypts the masterkey with the given user's hash and SID.
        Simply computes the corresponding key then calls self.decryptWithKey()
        """
        #print("Debug: Inside decryptwithhash10. userSID: "+userSID)
        #print("Debug: Inside decryptwithhash. pwdhash: "+binascii.hexlify(pwdhash))
        pwdhash1 = hashlib.pbkdf2_hmac('sha256', pwdhash, userSID.encode("UTF-16LE"), 10000)
        #print("Debug: Inside decryptwithhash. pwdhash1:"+binascii.hexlify(pwdhash1))
        pwdhash2 = hashlib.pbkdf2_hmac('sha256', pwdhash1, userSID.encode("UTF-16LE"), 1)[0:16]
        #print("Debug: Inside decryptwithhash. pwdhash2:"+binascii.hexlify(pwdhash2))
        
        #print("Debug: Inside decryptwithhash. Derived hash:" + binascii.hexlify(derivedkey))
        self.decryptWithKey(crypto.derivePwdHash(pwdhash2, userSID))

    def decryptWithPassword(self, userSID, pwd):
        """Decrypts the masterkey with the given user's password and SID.
        Simply computes the corresponding key, then calls self.decryptWithKey()

        """
        for algo in ["sha1", "md4"]:
            self.decryptWithKey(crypto.derivePwdHash(hashlib.new(algo, pwd.encode("UTF-16LE")).digest(), userSID))
            if self.decrypted:
                break

    def setKeyHash(self, h):
        assert(len(h) == 20)
        self.decrypted = True
        self.key_hash = h

    def setDecryptedKey(self, data):
        assert len(data) == 64
        self.decrypted = True
        self.key = data
        self.key_hash = hashlib.sha1(data).digest()

    def decryptWithKey(self, pwdhash):
        """Decrypts the masterkey with the given encryption key. This function
        also extracts the HMAC part of the decrypted stuff and compare it with
        the computed one.

        Note that, once successfully decrypted, the masterkey will not be
        decrypted anymore; this function will simply return.

        """
        if self.decrypted:
            return
        if not self.ciphertext:
            return
        # Compute encryption key
        cleartxt = crypto.dataDecrypt(self.cipherAlgo, self.hashAlgo, self.ciphertext,
                                      pwdhash, self.iv, self.rounds)
        self.key = cleartxt[-64:]
        self.hmacSalt = cleartxt[:16]
        self.hmac = cleartxt[16:16 + int(self.hashAlgo.digestLength)]
        self.hmacComputed = crypto.DPAPIHmac(self.hashAlgo, pwdhash, self.hmacSalt, self.key)
        self.decrypted = self.hmac == self.hmacComputed
        if self.decrypted:
            self.key_hash = hashlib.sha1(self.key).digest()

    def __repr__(self):
        s = ["Masterkey block"]
        if self.cipherAlgo is not None:
            s.append("\tcipher algo  = %s" % repr(self.cipherAlgo))
        if self.hashAlgo is not None:
            s.append("\thash algo    = %s" % repr(self.hashAlgo))
        if self.rounds is not None:
            s.append("\trounds       = %i" % self.rounds)
        if self.iv is not None:
            s.append("\tIV           = %s" % self.iv.hex())
        if self.key is not None:
            s.append("\tkey          = %s" % self.key.hex())
        if self.hmacSalt is not None:
            s.append("\thmacSalt     = %s" % self.hmacSalt.hex())
        if self.hmac is not None:
            s.append("\thmac         = %s" % self.hmac.hex())
        if self.hmacComputed is not None:
            s.append("\thmacComputed = %s" % self.hmacComputed.hex())
        if self.key_hash is not None:
            s.append("\tkey hash     = %s" % self.key_hash.hex())
        if self.ciphertext is not None:
            s.append("\tciphertext   = %s" % self.ciphertext.hex())
        return "\n".join(s)


class CredHist(eater.DataStruct):
    """This class represents a Credhist block contained in the MasterKeyFile"""

    def __init__(self, raw=None):
        self.version = None
        self.guid = None
        eater.DataStruct.__init__(self, raw)

    def parse(self, data):
        self.version = data.eat("L")
        self.guid = "%0x-%0x-%0x-%0x%0x-%0x%0x%0x%0x%0x%0x" % data.eat("L2H8B")

    def __repr__(self):
        s = ["CredHist block",
             "\tversion = %d" % self.version,
             "\tguid    = %s" % self.guid]
        return "\n".join(s)


class DomainKey(eater.DataStruct):
    """This class represents a DomainKey block contained in the MasterKeyFile.
    Currently does nothing more than parsing. Work on Active Directory stuff is
    still on progress.
    """
    def __init__(self, raw=None):
        self.version = None
        self.secretLen = None
        self.accesscheckLen = None
        self.guidKey = None
        self.encryptedSecret = None
        self.accessCheck = None
        self.decrypted = False
        self.key = None
        eater.DataStruct.__init__(self, raw)

    def parse(self, data):
        self.version = data.eat("L")
        self.secretLen = data.eat("L")
        self.accesscheckLen = data.eat("L")
        self.guidKey = "%0x-%0x-%0x-%0x%0x-%0x%0x%0x%0x%0x%0x" % data.eat("L2H8B")  #data.eat("16s")
        self.encryptedSecret = data.eat("%us" % self.secretLen)
        self.accessCheck = data.eat("%us" % self.accesscheckLen)

    def __repr__(self):
        s = ["DomainKey block",
             "\tversion     = %x" % self.version,
             "\tguid        = %s" % self.guidKey,
             "\tsecret      = %s" % self.encryptedSecret.hex(),
             "\taccessCheck = %s" % self.accessCheck.hex()]
        return "\n".join(s)

    def decryptWithDCKey(self, dcKeyFile):
        tmpdata = bytearray(self.encryptedSecret); tmpdata.reverse(); revmk=tmpdata
        rsakey = RSA.importKey(open(dcKeyFile, "r").read())
        cipher = PKCS1_v1_5.new(rsakey)
        decrypted_data = cipher.decrypt(binascii.unhexlify(binascii.hexlify(revmk)),0)
        if len(decrypted_data) > 0:
            self.key = decrypted_data[8:72]
            if self.key is not None:
                self.decrypted = True

    def decryptWithPVKKey(self, pvkKeyFile):
        def toInt(bHex):
            return int.from_bytes(bHex, 'little')
        ## First 6 DWORDs are ignored
        bData1 = open(pvkKeyFile,'rb').read()[6*4:]
        ## RSAStruct starts with magic bytes 'RSA2'
        bRSAStruct = bData1[2*4:]
        iBitLength1 = int(toInt(bRSAStruct[4:2*4]) / 8)
        iBitLength2 = int(iBitLength1 / 2)
        iPubExp = toInt(bRSAStruct[2*4:3*4])
        ## After 'RSA2', bitlength and Public Exponent: the private key data
        bPrivKeyStruct = bRSAStruct[3*4:]
        iModulus = toInt(bPrivKeyStruct[:iBitLength1])
        iPrime1 = toInt(bPrivKeyStruct[iBitLength1:iBitLength1+iBitLength2])
        iPrime2 = toInt(bPrivKeyStruct[iBitLength1+iBitLength2:iBitLength1+iBitLength2+iBitLength2])
        iPrivExp = toInt(bPrivKeyStruct[int(-1*iBitLength1):])
        rsakey = RSA.construct((iModulus,iPubExp,iPrivExp,iPrime1,iPrime2))
        cipher = PKCS1_v1_5.new(rsakey)
        decrypted_data = cipher.decrypt(self.encryptedSecret[::-1],0)
        if not decrypted_data == 0:
            self.key = decrypted_data[8:72]
            if self.key is not None:
                self.decrypted = True
        

class MasterKeyFile(eater.DataStruct):
    """This class represents a masterkey file."""

    def __init__(self, raw=None):
        self.masterkey = None
        self.backupkey = None
        self.credhist = None
        self.domainkey = None
        self.decrypted = False
        self.version = None
        self.guid = None
        self.policy = None
        self.masterkeyLen = self.backupkeyLen = self.credhistLen = self.domainkeyLen = 0
        eater.DataStruct.__init__(self, raw)

    def parse(self, data):
        self.version = data.eat("L")
        data.eat("2L")
        self.guid = data.eat("72s").decode("UTF-16LE").encode("utf-8")
        data.eat("2L")
        self.policy = data.eat("L")
        self.masterkeyLen = data.eat("Q")
        self.backupkeyLen = data.eat("Q")
        self.credhistLen = data.eat("Q")
        self.domainkeyLen = data.eat("Q")

        if self.masterkeyLen > 0:
            self.masterkey = MasterKey()
            self.masterkey.parse(data.eat_sub(self.masterkeyLen))
        if self.backupkeyLen > 0:
            self.backupkey = MasterKey()
            self.backupkey.parse(data.eat_sub(self.backupkeyLen))
        if self.credhistLen > 0:
            self.credhist = CredHist()
            self.credhist.parse(data.eat_sub(self.credhistLen))
        if self.domainkeyLen > 0:
            try:
                self.domainkey = DomainKey()
                self.domainkey.parse(data.eat_sub(self.domainkeyLen))
            except:
                print('Error with reading domain Key')
                self.domainkeyLen = 0
                self.domainkey = None

    def decryptWithHash(self, userSID, h, alg='sha1'):
        """See MasterKey.decryptWithHash()"""
        if not self.masterkey.decrypted:
            self.masterkey.decryptWithHash10(userSID, h)
            if not self.masterkey.decrypted:
                self.masterkey.decryptWithHash(userSID, h)
        #disabling backup key decrypting for more quickly....
        #if not self.backupkey.decrypted and alg =='sha1':
        #    self.backupkey.decryptWithHash(userSID, h)
        #if not self.backupkey.decrypted and alg=='md4':
        #    self.backupkey.decryptWithHash(userSID, h)
        #    if not self.backupkey.decrypted:
        #        self.backupkey.decryptWithHash10(userSID, h)
        self.decrypted = self.masterkey.decrypted or self.backupkey.decrypted

    def decryptWithPassword(self, userSID, pwd):
        """See MasterKey.decryptWithPassword()"""
        for algo in ["sha1", "md4"]:
            self.decryptWithHash(userSID, hashlib.new(algo, pwd.encode('UTF-16LE')).digest(), algo)
            
            if self.decrypted: break

    def decryptWithKey(self, pwdhash):
        """See MasterKey.decryptWithKey()"""
        if not self.masterkey.decrypted:
            self.masterkey.decryptWithKey(pwdhash)
        if not self.backupkey.decrypted:
            self.backupkey.decryptWithKey(pwdhash)
        self.decrypted = self.masterkey.decrypted or self.backupkey.decrypted

    def addKeyHash(self, guid, h):
        self.guid = guid
        self.masterkey = MasterKey()
        self.backupkey = MasterKey()
        self.masterkey.setKeyHash(h)
        self.decrypted = True

    def addDecryptedKey(self, guid, data):
        self.guid = guid
        self.masterkey = MasterKey()
        self.backupkey = MasterKey()
        self.masterkey.setDecryptedKey(data)
        self.decrypted = True

    def get_key(self):
        """Returns the first decrypted block between Masterkey and BackupKey.
        If none has been decrypted, returns the Masterkey block.
        """
        if self.masterkey.decrypted:
            return self.masterkey.key or self.masterkey.key_hash
        elif self.backupkey.decrypted:
            return self.backupkey.key
        return self.masterkey.key

    def __repr__(self):
        s = ["\n#### MasterKeyFile %s ####" % self.guid.decode()]
        if self.version is not None:
            s.append("\tversion   = %#d" % self.version)
        if self.policy is not None:
            s.append("\tPolicy    = %#x" % self.policy)
        if self.masterkeyLen > 0:
            s.append("\tMasterKey = %d" % self.masterkeyLen)
        if self.backupkeyLen > 0:
            s.append("\tBackupKey = %d" % self.backupkeyLen)
        if self.credhistLen > 0:
            s.append("\tCredHist  = %d" % self.credhistLen)
        if self.domainkeyLen > 0:
            s.append("\tDomainKey = %d" % self.domainkeyLen)
        if self.masterkey:
            s.append("    + Master Key: %s" % repr(self.masterkey))
        if self.backupkey:
            s.append("    + Backup Key: %s" % repr(self.backupkey))
        if self.credhist:
            s.append("    + %s" % repr(self.credhist))
        if self.domainkey:
            s.append("    + %s" % repr(self.domainkey))
        return "\n".join(s)


class MasterKeyPool(object):
    """This class is the pivot for using DPAPIck. It manages all the DPAPI
    structures and contains all the decryption intelligence.
    """

    def __init__(self):
        self.keys = defaultdict(lambda: [])
        self.creds = {}
        self.system = None
        self.passwords = set()

    def addMasterKey(self, mkey):
        """Add a MasterKeyFile is the pool.
        mkey is a string representing the content of the file to add.
        """
        mkf = MasterKeyFile(mkey)
        self.keys[mkf.guid].append(mkf)

    def addMasterKeyHash(self, guid, h):
        self.keys[guid].append(MasterKeyFile().addKeyHash(guid, h))

    def getMasterKeys(self, guid):
        """Returns an array of Masterkeys corresponding the the given GUID.
        """
        return self.keys.get(guid, [])

    def addSystemCredential(self, blob):
        """Adds DPAPI_SYSTEM token to the pool.
        blob is a string representing the LSA secret token
        """
        self.system = credhist.CredSystem(blob)

    def addCredhist(self, sid, cred):
        """Internal use. Adds a CredHistFile to the pool.
        sid is a string representing the user's SID
        cred is CredHistFile object.
        """
        self.creds[sid] = cred

    def addCredhistFile(self, sid, credfile):
        """Adds a Credhist file to the pool.
        sid is a string representing the user's SID
        credfile is the full path to the CREDHIST file to add.
        """
        with open(credfile, 'rb') as f:
            self.addCredhist(sid, credhist.CredHistFile(f.read()))

    def loadDirectory(self, directory):
        """Adds every masterkey contained in the given directory to the pool.
        If a file is not a valid Masterkey file, this function simply goes to
        the next file without complaining.

        directory is a string representing the directory path to add.
        """
        for k in os.listdir(directory):
            if re.match("^[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}$", k, re.IGNORECASE):
                try:
                    with open(os.path.join(directory, k), 'rb') as f:
                        self.addMasterKey(f.read())
                except:
                    pass

    def pickle(self, filename=None):
        if filename is not None:
            pickle.dump(self, filename, 2)
        else:
            return pickle.dumps(self, 2)

    def __getstate__(self):
        d = dict(self.__dict__)
        d["keys"] = dict(d["keys"])
        return d

    def __setstate__(self, d):
        tmp = dict(d["keys"])
        d["keys"] = defaultdict(lambda: [])
        d["keys"].update(tmp)
        self.__dict__.update(d)

    @staticmethod
    def unpickle(data=None, filename=None):
        if data is not None:
            return pickle.loads(data)
        if filename is not None:
            return pickle.load(filename)
        raise ValueError("must provide either data or filename argument")

    def try_credential_hash(self, userSID, pwdhash):
        n = 0
        for mkl in list(self.keys.values()):
            for mk in mkl:
                if not mk.decrypted:
                    if pwdhash is not None:
                        mk.decryptWithHash(userSID, pwdhash)
                        if not mk.decrypted and self.creds.get(userSID) is not None:
                            # process CREDHIST
                            self.creds[userSID].decryptWithHash(pwdhash)
                            for cred in self.creds[userSID].entries_list:
                                mk.decryptWithHash(userSID, cred.pwdhash)
                                if cred.ntlm is not None and not mk.decrypted:
                                    mk.decryptWithHash(userSID, cred.ntlm)
                                if mk.decrypted:
                                    self.creds[userSID].validate()
                                    break
                    if not mk.decrypted and self.system is not None:
                        # try DPAPI_SYSTEM creds
                        mk.decryptWithKey(self.system.user)
                        if not mk.decrypted:
                            mk.decryptWithKey(self.system.machine)
                        if userSID is not None and not mk.decrypted:
                            # try with an extra SID (just in case)
                            mk.decryptWithHash(userSID, self.system.user)
                            if not mk.decrypted:
                                mk.decryptWithHash(userSID, self.system.machine)
                    if mk.decrypted:
                        n += 1
        return n

    def try_credential(self, userSID, password):
        """This function tries to decrypt every masterkey contained in the pool
        that has not been successfully decrypted yet with the given password and
        SID.

        userSID is a string representing the user's SID
        password is a string representing the user's password.

        Returns the number of masterkey that has been successfully decrypted
        with those credentials.

        """
        n = 0
        for mkl in list(self.keys.values()):
            for mk in mkl:
                if not mk.decrypted:
                    if password is not None:
                        mk.decryptWithPassword(userSID, password)
                        if not mk.decrypted and self.creds.get(userSID) is not None:
                            # process CREDHIST
                            self.creds[userSID].decryptWithPassword(password)
                            for cred in self.creds[userSID].entries_list:
                                mk.decryptWithHash(userSID, cred.pwdhash)
                                if cred.ntlm is not None and not mk.decrypted:
                                    mk.decryptWithHash(userSID, cred.ntlm)
                                if mk.decrypted:
                                    self.creds[userSID].validate()
                                    break
                    if not mk.decrypted and self.system is not None:
                        # try DPAPI_SYSTEM creds
                        mk.decryptWithHash(userSID, self.system.user)
                        if not mk.decrypted:
                            mk.decryptWithHash(userSID, self.system.machine)
                        if not mk.decrypted:
                            mk.decryptWithKey(self.system.user)
                        if not mk.decrypted:
                            mk.decryptWithKey(self.system.machine)
                    if mk.decrypted:
                        self.passwords.add(password)
                        n += 1
        return n

    def try_domain(self, privkeyfile):
        """
        try to decrypt domain backup key with DC RSA private key
        """
        n = 0
        for mkl in list(self.keys.values()):
            for mk in mkl:
                if not mk.decrypted and mk.domainkey:
                    try: mk.domainkey.decryptWithDCKey(privkeyfile)
                    except: mk.domainkey.decryptWithPVKKey(privkeyfile)
                    if mk.domainkey.decrypted:
                        mk.decrypted = True
                        mk.masterkey.key = mk.domainkey.key

                    if mk.decrypted:
                        n += 1
        return n        


    def __repr__(self):
        #s = ['MasterKeyPool:', 'Passwords:', repr(self.passwords), 'Keys:', repr(self.keys.items())]
        s = ['MasterKeyPool:', 'Passwords:', repr(self.passwords), 'Keys:', repr(list(self.keys.items()))]
        if self.system is not None:
            s.append(repr(self.system))
        s.append("CredHist entries:")
        #for i in self.creds.keys():
        for i in list(self.creds.keys()):
            s.append("\tSID: %s" % i)
            s.append(repr(self.creds[i]))
        return "\n".join(s)

# vim:ts=4:expandtab:sw=4



#!/usr/bin/python3
# -*- coding: utf-8 -*-

#############################################################################
##                                                                         ##
## This file is part of DPAPIck                                            ##
## Windows DPAPI decryption & forensic toolkit                             ##
##                                                                         ##
##                                                                         ##
## Copyright (C) 2010, 2011 Cassidian SAS. All rights reserved.            ##
## Copyright (C) 2023       Insecurity. All rights reserved.               ##
##                                                                         ##
##  Author:  Jean-Michel Picod <jmichel.p@gmail.com>                       ##
##  Updated: Photubias <info@insecurity.be>                                ##
##                                                                         ##
## This program is distributed under GPLv3 licence (see LICENCE.txt)       ##
##                                                                         ##
#############################################################################

from dpapick3 import probe, blob
from Crypto.Math.Numbers import Integer
from Crypto.PublicKey import RSA

try:
    from pyasn1.type import univ, namedtype
    from pyasn1.codec.der import encoder
except:
    raise Exception("PyASN1 is required.")

class PrivateKeyBlob(probe.DPAPIProbe):
    """This class represents a RSA private key file as used by Internet Explorer
    and EFS certificates.
    They are located under %APPDATA%\\Microsoft\\Crypto\\RSA
    """

    class RSAHeader(probe.DPAPIProbe):
        """This subclass represents the header + modulus, beginning with the
        magic value "RSA1"
        """
        def parse(self, data):
            self.magic = data.eat("4s")  # RSA1
            self.len1 = data.eat("L")  # e.g. 0x108 (264)
            self.bitlength = data.eat("L")  # e.g. 0x400 (1024) or 0x800 (2048)
            self.unk = data.eat("L")  # e.g. 0x7F (127) or 0xFF (255)
            self.pubexp = data.eat("L")  # e.g. 0x00010001 (65537)
            self.data = data.eat("%is" % self.len1)
            self.data = self.data.strip(b'\x00') # strip NULL-bytes

        def __repr__(self):
            s = ["RSA header",
                 "\tmagic     = %s" % self.magic,
                 "\tbitlength = %d" % self.bitlength,
                 "\tunknown   = %x" % self.unk,
                 "\tpubexp    = %d" % self.pubexp,
                 "\tdata      = %s" % self.data.hex()]
            return "\n".join(s)

    class RSAKey(probe.DPAPIProbe):
        """This subclass represents the RSA privatekey BLOB, beginning with the
        magic value "RSA2"
        """

        class RSAKeyASN1(univ.Sequence):
            """subclass for ASN.1 sequence representing the RSA key pair.
            Mainly useful to export the key to OpenSSL
            """
            componentType = namedtype.NamedTypes(
                namedtype.NamedType('version', univ.Integer()),
                namedtype.NamedType('modulus', univ.Integer()),
                namedtype.NamedType('pubexpo', univ.Integer()),
                namedtype.NamedType('privexpo', univ.Integer()),
                namedtype.NamedType('prime1', univ.Integer()),
                namedtype.NamedType('prime2', univ.Integer()),
                namedtype.NamedType('exponent1', univ.Integer()),
                namedtype.NamedType('exponent2', univ.Integer()),
                namedtype.NamedType('coefficient', univ.Integer())
            )

        def parse(self, data):
            self.magic = data.eat("4s")  # RSA2
            len1 = data.eat("L")
            self.bitlength = data.eat("L")
            chunk = int(self.bitlength / 16)
            self.unk = data.eat("L")
            self.pubexp = data.eat("L")
            self.modulus = data.eat("%is" % len1)[:2 * chunk]
            self.prime1 = data.eat("%is" % (len1 / 2))[:chunk]
            self.prime2 = data.eat("%is" % (len1 / 2))[:chunk]
            self.exponent1 = data.eat("%is" % (len1 / 2))[:chunk]
            self.exponent2 = data.eat("%is" % (len1 / 2))[:chunk]
            self.coefficient = data.eat("%is" % (len1 / 2))[:chunk]
            self.privExponent = data.eat("%is" % len1)[:2 * chunk]
            self.asn1 = self.RSAKeyASN1()
            #ll = lambda x: long(x[::-1].hex(), 16)
            ll = lambda x: int(x[::-1].hex(), 16)
            self.asn1.setComponentByName('version', 0)
            self.asn1.setComponentByName('modulus', ll(self.modulus))
            self.asn1.setComponentByName('pubexpo', self.pubexp)
            self.asn1.setComponentByName('privexpo', ll(self.privExponent))
            self.asn1.setComponentByName('prime1', ll(self.prime1))
            self.asn1.setComponentByName('prime2', ll(self.prime2))
            self.asn1.setComponentByName('exponent1', ll(self.exponent1))
            self.asn1.setComponentByName('exponent2', ll(self.exponent2))
            self.asn1.setComponentByName('coefficient', ll(self.coefficient))

        def __repr__(self):
            s = ["RSA key pair",
                 "\tPublic exponent = %d" % self.pubexp,
                 "\tModulus (n)     = %s" % self.modulus.hex(),
                 "\tPrime 1 (p)     = %s" % self.prime1.hex(),
                 "\tPrime 2 (q)     = %s" % self.prime2.hex(),
                 "\tExponent 1      = %s" % self.exponent1.hex(),
                 "\tExponent 2      = %s" % self.exponent2.hex(),
                 "\tCoefficient     = %s" % self.coefficient.hex(),
                 "\tPrivate exponent= %s" % self.privExponent.hex()]
            return "\n".join(s)

        def export(self):
            """This functions exports the RSA key pair in PEM format"""
            import base64
            s = ['-----BEGIN RSA PRIVATE KEY-----']
            text = base64.b64encode(encoder.encode(self.asn1)).decode()
            s.append(text.rstrip('\n'))
            s.append('-----END RSA PRIVATE KEY-----')
            return '\n'.join(s)

    class RSAPrivKey(probe.DPAPIProbe):
        """Internal use. This represents the DPAPI BLOB containing the RSA
        key pair"""
        def parse(self, data):
            self.dpapiblob = blob.DPAPIBlob(data.remain())
            self.dpapiblob.guids = [self.dpapiblob.mkguid.encode()]

        def postprocess(self, **k):
            self.clearKey = PrivateKeyBlob.RSAKey(self.dpapiblob.cleartext)

        def export(self):
            if self.clearKey is None: return ''
            return self.clearKey.export()
        
        def rsa(self):
            n = Integer.from_bytes(self.clearKey.modulus[::-1])
            p = Integer.from_bytes(self.clearKey.prime1[::-1])
            q = Integer.from_bytes(self.clearKey.prime2[::-1])
            e = Integer(self.clearKey.pubexp)
            lcm = (p - 1).lcm(q - 1)
            d = e.inverse(lcm)
            return RSA.construct((int(n), int(e), int(d), int(p), int(q)))

        def __repr__(self):
            s = ["RSA Private Key Blob"]
            if self.entropy:
                s.append("entropy = %s" % self.entropy.hex())
            if hasattr(self, "strong"):
                s.append("strong = %s" % self.strong.hex())
            if self.dpapiblob.decrypted:
                s.append(repr(self.clearKey))
            s.append(repr(self.dpapiblob))
            return "\n".join(s)

    class RSAFlags(probe.DPAPIProbe):
        """This subclass represents the export flags BLOB"""
        def parse(self, data):
            self.dpapiblob = blob.DPAPIBlob(data.remain())
            self.dpapiblob.guids = [self.dpapiblob.mkguid.encode()]

        def preprocess(self, **k):
            self.entropy = b"Hj1diQ6kpUx7VC4m\0"
            if hasattr(k, 'strong'):
                self.strong = k['strong']

        def __repr__(self):
            s = ["Export Flags"]
            s.append("entropy = %s" % self.entropy)
            if hasattr(self, "strong"):
                s.append("strong = %s" % self.strong.encode("hex"))
            s.append("%r" % self.dpapiblob)
            return "\n".join(s)

    def parse(self, data):
        self.version = data.eat("L")
        data.eat("L")  # NULL
        self.descrLen = data.eat("L")
        sigheadlen, sigprivkeylen = data.eat("2L")
        headerlen = data.eat("L")
        privkeylen = data.eat("L")
        self.crcLen = data.eat("L")
        sigflagslen = data.eat("L")
        flagslen = data.eat("L")

        if headerlen == 0:
            headerlen = sigheadlen
            sigheadlen = 0
        if privkeylen == 0:
            privkeylen = sigprivkeylen
            sigprivkeylen = 0
        if flagslen == 0:
            flagslen = sigflagslen
            sigflagslen = 0


        self.description = data.eat("%ds" % self.descrLen)
        self.description = self.description.strip(b'\x00')
        self.crc = data.eat("%ds" % self.crcLen)

        # Signature key comes first
        self.sigHeader = None
        if sigheadlen > 0:
            self.sigHeader = self.RSAHeader()
            self.sigHeader.parse(data.eat_sub(sigheadlen))

        self.sigPrivateKey = None
        if sigprivkeylen > 0:
            self.sigPrivateKey = self.RSAPrivKey()
            self.sigPrivateKey.parse(data.eat_sub(sigprivkeylen))

        self.sigFlags = None
        if sigflagslen > 0:
            self.sigFlags = self.RSAFlags()
            self.sigFlags.parse(data.eat_sub(sigflagslen))

        # Then export key
        self.header = None
        if headerlen > 0:
            self.header = self.RSAHeader()
            self.header.parse(data.eat_sub(headerlen))

        self.privateKey = None
        if privkeylen > 0:
            self.privateKey = self.RSAPrivKey()
            self.privateKey.parse(data.eat_sub(privkeylen))

        self.flags = None
        if flagslen > 0:
            self.flags = self.RSAFlags()
            self.flags.parse(data.eat_sub(flagslen))

    def try_decrypt_with_hash(self, h, mkp, sid, **k):
        if not self.flags:
            return False
        if not self.privateKey:
            return False
        if self.flags.try_decrypt_with_hash(h, mkp, sid, **k):
            self.privateKey.entropy = self.flags.cleartext
            return self.privateKey.try_decrypt_with_hash(h, mkp, sid, **k)
        return False

    def try_decrypt_with_password(self, password, mkp, sid, **k):
        if not self.flags:
            return False
        if not self.privateKey:
            return False
        if self.flags.try_decrypt_with_password(password, mkp, sid, **k):
            self.privateKey.entropy = self.flags.cleartext
            return self.privateKey.try_decrypt_with_password(password, mkp, sid, **k)
        return False
    
    def try_decrypt_with_key(self, key, **k):
        if not self.flags: return False
        if not self.privateKey: return False
        if self.flags.try_decrypt_with_key(key, **k):
            self.privateKey.entropy = self.flags.cleartext
            return self.privateKey.try_decrypt_with_key(key, **k)
        return False

    def export(self):
        """This functions encodes the RSA key pair in PEM format. Simply calls the same function on the key blob."""
        if self.privateKey:
            return self.privateKey.export()
        return ''

    def __repr__(self):
        s = ["Microsoft Certificate",
             "\tdescr: %s" % self.description]
        if self.header is not None:
            s.append("+  %r" % self.header)
        if self.privateKey is not None:
            s.append("+  %r" % self.privateKey)
        if self.flags is not None:
            s.append("+  %r" % self.flags)
        return "\n".join(s)

class BcryptPrivateKeyBlob(probe.DPAPIProbe):
    """This class represents a Bcrypt RSA private key blob file 
    They are located under %APPDATA%\\Microsoft\\Crypto\\Keys
    """

    class RSAHeader(probe.DPAPIProbe):
        """This subclass represents the header + modulus, beginning with the
        magic value "RSA1"
        """
        def parse(self, data):
            self.magic = data.eat("4s")  # RSA1
            self.len1 = data.eat("L")  # e.g. 0x108 (264)
            self.bitlength = data.eat("L")  # e.g. 0x400 (1024) or 0x800 (2048)
            self.unk = data.eat("L")  # e.g. 0x7F (127) or 0xFF (255)
            self.pubexp = data.eat("L")  # e.g. 0x00010001 (65537)
            self.data = data.eat("%is" % self.len1)
            self.data = self.data.strip(b'\x00') # strip NULL-bytes

        def __repr__(self):
            s = ["RSA header",
                 "\tmagic     = %s" % self.magic,
                 "\tbitlength = %d" % self.bitlength,
                 "\tunknown   = %x" % self.unk,
                 "\tpubexp    = %d" % self.pubexp,
                 "\tdata      = %s" % self.data.hex()]
            return "\n".join(s)

    class RSAKey(probe.DPAPIProbe):
        """This subclass represents the RSA privatekey BLOB, beginning with the
        magic value "RSA2"
        """

        class RSAKeyASN1(univ.Sequence):
            """subclass for ASN.1 sequence representing the RSA key pair.
            Mainly useful to export the key to OpenSSL
            """
            componentType = namedtype.NamedTypes(
                namedtype.NamedType('version', univ.Integer()),
                namedtype.NamedType('modulus', univ.Integer()),
                namedtype.NamedType('pubexpo', univ.Integer()),
                # namedtype.NamedType('privexpo', univ.Integer()),
                namedtype.NamedType('prime1', univ.Integer()),
                namedtype.NamedType('prime2', univ.Integer()),
                # namedtype.NamedType('exponent1', univ.Integer()),
                # namedtype.NamedType('exponent2', univ.Integer()),
                # namedtype.NamedType('coefficient', univ.Integer())
            )

        def parse(self, data):
            self.magic = data.eat("4s")  # RSA2
            byte_len = data.eat("L")
            pub_exp_len = data.eat("L")
            modulus_len = data.eat("L")
            p_len = data.eat("L")
            q_len = data.eat("L")
            self.pubexp = data.eat("%is" % pub_exp_len)
            self.modulus = data.eat("%is" % modulus_len)
            self.prime1 = data.eat("%is" % p_len)
            self.prime2 = data.eat("%is" % q_len)

            # self.privExponent = self.pubexp
            # self.exponent1 = self.pubexp
            # self.exponent2 = self.pubexp
            # self.coefficient = self.pubexp

            self.asn1 = self.RSAKeyASN1()
            # ll = lambda x: long(x[::-1].hex(), 16)
            ll = lambda x: int(x[::-1].hex(), 16)
            self.asn1.setComponentByName('version', 0)
            self.asn1.setComponentByName('modulus', ll(self.modulus))
            self.asn1.setComponentByName('pubexpo', ll(self.pubexp))
            # self.asn1.setComponentByName('privexpo', ll(self.privExponent))
            self.asn1.setComponentByName('prime1', ll(self.prime1))
            self.asn1.setComponentByName('prime2', ll(self.prime2))
            # self.asn1.setComponentByName('exponent1', ll(self.exponent1))
            # self.asn1.setComponentByName('exponent2', ll(self.exponent2))
            # self.asn1.setComponentByName('coefficient', ll(self.coefficient))

        def __repr__(self):
            s = ["RSA key pair",
                 "\tPublic exponent = %d" % self.pubexp,
                 "\tModulus (n)     = %s" % self.modulus.hex(),
                 "\tPrime 1 (p)     = %s" % self.prime1.hex(),
                 "\tPrime 2 (q)     = %s" % self.prime2.hex()]
            # "\tExponent 1      = %s" % self.exponent1.hex(),
            # "\tExponent 2      = %s" % self.exponent2.hex(),
            # "\tCoefficient     = %s" % self.coefficient.hex(),
            # "\tPrivate exponent= %s" % self.privExponent.hex()]
            return "\n".join(s)

        def export(self):
            """This functions exports the RSA key pair in PEM format"""
            import base64
            s = ['-----BEGIN RSA PRIVATE KEY-----']
            text = base64.b64encode(encoder.encode(self.asn1)).decode()
            s.append(text.rstrip('\n'))
            s.append('-----END RSA PRIVATE KEY-----')
            return '\n'.join(s)

        def export_pkcs12(self):
            n = Integer.from_bytes(self.modulus)
            p = Integer.from_bytes(self.prime1)
            q = Integer.from_bytes(self.prime2)
            e = Integer.from_bytes(self.pubexp)
            lcm = (p - 1).lcm(q - 1)
            d = e.inverse(lcm)
            rsa_components = (int(n), int(e), int(d))
            return RSA.construct(rsa_components).export_key()

    class RSAPrivKey(probe.DPAPIProbe):
        """Internal use. This represents the DPAPI BLOB containing the RSA
        key pair"""
        def parse(self, data):
            self.dpapiblob = blob.DPAPIBlob(data.remain())

        def postprocess(self, **k):
            self.clearKey = BcryptPrivateKeyBlob.RSAKey(self.dpapiblob.cleartext)

        def export(self):
            if self.clearKey is None:
                return ""
            return self.clearKey.export()

        def __repr__(self):
            s = ["RSA Private Key Blob"]
            if self.entropy:
                s.append("entropy = %s" % self.entropy.hex())
            if hasattr(self, "strong"):
                s.append("strong = %s" % self.strong.hex())
            if self.dpapiblob.decrypted:
                s.append(repr(self.clearKey))
            s.append(repr(self.dpapiblob))
            return "\n".join(s)

    class RSAFlags(probe.DPAPIProbe):
        """This subclass represents the export flags BLOB"""
        def parse(self, data):
            self.dpapiblob = blob.DPAPIBlob(data.remain())

        def preprocess(self, **k):
            self.entropy = b"Hj1diQ6kpUx7VC4m\0"
            if hasattr(k, 'strong'):
                self.strong = k['strong']

        def __repr__(self):
            s = ["Export Flags"]
            s.append("entropy = %s" % self.entropy)
            if hasattr(self, "strong"):
                s.append("strong = %s" % self.strong.encode("hex"))
            s.append("%r" % self.dpapiblob)
            return "\n".join(s)

    def parse(self, data):
        self.version = data.eat("L")
        data.eat("L")  # NULL
        self.descrLen = data.eat("L")
        sigheadlen, sigprivkeylen = data.eat("2L")
        headerlen = data.eat("L")
        privkeylen = data.eat("L")
        self.crcLen = data.eat("L")
        sigflagslen = data.eat("L")
        flagslen = data.eat("L")

        if headerlen == 0:
            headerlen = sigheadlen
            sigheadlen = 0
        if privkeylen == 0:
            privkeylen = sigprivkeylen
            sigprivkeylen = 0
        if flagslen == 0:
            flagslen = sigflagslen
            sigflagslen = 0


        self.description = data.eat("%ds" % self.descrLen)
        self.description = self.description.strip(b'\x00')
        self.crc = data.eat("%ds" % self.crcLen)

        # Signature key comes first
        self.sigHeader = None
        if sigheadlen > 0:
            self.sigHeader = self.RSAHeader()
            self.sigHeader.parse(data.eat_sub(sigheadlen))

        self.sigPrivateKey = None
        if sigprivkeylen > 0:
            self.sigPrivateKey = self.RSAPrivKey()
            self.sigPrivateKey.parse(data.eat_sub(sigprivkeylen))

        self.sigFlags = None
        if sigflagslen > 0:
            self.sigFlags = self.RSAFlags()
            self.sigFlags.parse(data.eat_sub(sigflagslen))

        # Then export key
        self.header = None
        if headerlen > 0:
            self.header = self.RSAHeader()
            self.header.parse(data.eat_sub(headerlen))

        self.privateKey = None
        if privkeylen > 0:
            self.privateKey = self.RSAPrivKey()
            self.privateKey.parse(data.eat_sub(privkeylen))

        self.flags = None
        if flagslen > 0:
            self.flags = self.RSAFlags()
            self.flags.parse(data.eat_sub(flagslen))

    def try_decrypt_with_hash(self, h, mkp, sid, **k):
        if not self.flags:
            return False
        if not self.privateKey:
            return False
        if self.flags.try_decrypt_with_hash(h, mkp, sid, **k):
            self.privateKey.entropy = self.flags.cleartext
            return self.privateKey.try_decrypt_with_hash(h, mkp, sid, **k)
        return False

    def try_decrypt_with_password(self, password, mkp, sid, **k):
        if not self.flags:
            return False
        if not self.privateKey:
            return False
        if self.flags.try_decrypt_with_password(password, mkp, sid, **k):
            self.privateKey.entropy = self.flags.cleartext
            return self.privateKey.try_decrypt_with_password(password, mkp, sid, **k)
        return False
    
    def try_decrypt_with_key(self, key, **k):
        if not self.flags: return False
        if not self.privateKey: return False
        if self.flags.try_decrypt_with_key(key, **k):
            self.privateKey.entropy = self.flags.cleartext
            return self.privateKey.try_decrypt_with_key(key, **k)
        return False

    def export(self):
        """This functions encodes the RSA key pair in PEM format. Simply calls the same function on the key blob."""
        if self.privateKey:
            return self.privateKey.export()
        return ''

    def __repr__(self):
        s = ["Microsoft Certificate",
             "\tdescr: %s" % self.description]
        if self.header is not None:
            s.append("+  %r" % self.header)
        if self.privateKey is not None:
            s.append("+  %r" % self.privateKey)
        if self.flags is not None:
            s.append("+  %r" % self.flags)
        return "\n".join(s)

class Cert(probe.DPAPIProbe):
    PROP_PROVIDER = 2
    PROP_CERTIFICATE = 0x20

    PROPS = {2: "Provider", 3: "SHA-1", 4: "MD5", 15: "Signature hash",
             20: "Key identifier", 25: "Subject MD5 hash",
             69: "Backed up", 92: "Pubkey bitlength", 32: "ASN.1 certificate"}

    class CertProvider(probe.DPAPIProbe):
        def parse(self, data):
            raw = data.remain()
            ofs_keyname, ofs_provider = data.eat("2L")
            self.provider_type = data.eat("L")
            self.flags = data.eat("L")
            self.nb_params = data.eat("L")
            ofs_params = data.eat("L")
            self.key_specs = data.eat("L")
            self.keyname = raw[ofs_keyname:]
            end = self.keyname.find(b'\x00\x00\x00')
            if end & 1: end += 1
            self.keyname = self.keyname[:end].decode('UTF-16LE').rstrip('\x00')
            self.provider = raw[ofs_provider:]
            end = self.provider.find(b'\x00\x00\x00')
            if end & 1: end += 1
            self.provider = self.provider[:end].decode('UTF-16LE').rstrip('\x00')

        def __str__(self, indent=''):
            rv = ['%s+ Cert Provider' % indent]
            rv.append('%s  - Key name: %s' % (indent, self.keyname))
            rv.append('%s  - Provider: %s' % (indent, self.provider))
            return '\n'.join(rv)


    def parse(self, data):
        self.props = {}
        while data:
            cert_id, n, size = data.eat("3L")
            self.props[cert_id] = []
            for _ in range(n):
                if cert_id == self.PROP_PROVIDER:
                    provider = self.CertProvider()
                    provider.parse(data.eat_sub(size))
                    self.props[cert_id].append(provider)
                    self.name = provider.keyname
                elif cert_id == self.PROP_CERTIFICATE:
                    self.certificate = data.eat_string(size)
                else:
                    self.props[cert_id].append(data.eat_string(size))

    def __str__(self):
        rv = ["+ Microsoft Certificate"]
        for k, v in self.props.items():
            rv.append("  + %s (%d)" % (self.PROPS.get(k, "Unknown"), k))
            for p in v:
                if isinstance(p, str) or isinstance(p, bytes): rv.append('     - %s' % p.hex())
                else: rv.append(str(p))
        return '\n'.join(rv)

# vim:ts=4:expandtab:sw=4



#!/usr/bin/python3
# -*- coding: utf-8 -*-

#############################################################################
##                                                                         ##
## This file is part of DPAPIck                                            ##
## Windows DPAPI decryption & forensic toolkit                             ##
##                                                                         ##
##                                                                         ##
## Copyright (C) 2010, 2011 Cassidian SAS. All rights reserved.            ##
## Copyright (C) 2023       Insecurity. All rights reserved.               ##
##                                                                         ##
##  Author:  Jean-Michel Picod <jmichel.p@gmail.com>                       ##
##  Updated: Photubias <info@insecurity.be>                                ##
##                                                                         ##
## This program is distributed under GPLv3 licence (see LICENCE.txt)       ##
##                                                                         ##
#############################################################################

import dpapick3.crypto as crypto
import dpapick3.eater as eater

class DPAPIBlob(eater.DataStruct):
    """Represents a DPAPI blob"""

    def __init__(self, raw=None):
        """Constructs a DPAPIBlob. If raw is set, automatically calls
            parse().

        """
        self.version = None
        self.provider = None
        self.mkguid = None
        self.mkversion = None
        self.flags = None
        self.description = None
        self.cipherAlgo = None
        self.keyLen = 0
        self.hmac = None
        self.strong = None
        self.hashAlgo = None
        self.hashLen = 0
        self.cipherText = None
        self.salt = None
        self.blob = None
        self.sign = None
        self.cleartext = None
        self.decrypted = False
        self.signComputed = None
        eater.DataStruct.__init__(self, raw)

    def parse(self, data):
        """Parses the given data. May raise exceptions if incorrect data are
            given. You should not call this function yourself; DataStruct does

            data is a DataStruct object.
            Returns nothing.

        """
        self.version = data.eat("L")
        self.provider = "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x" % data.eat("L2H8B")

        # For HMAC computation
        blobStart = data.ofs

        self.mkversion = data.eat("L")
        self.mkguid = "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x" % data.eat("L2H8B")

        self.flags = data.eat("L")
        self.description = data.eat_length_and_string("L").decode("UTF-16LE").encode("utf-8")
        self.cipherAlgo = crypto.CryptoAlgo(data.eat("L"))
        self.keyLen = data.eat("L")
        self.salt = data.eat_length_and_string("L")
        self.strong = data.eat_length_and_string("L")
        self.hashAlgo = crypto.CryptoAlgo(data.eat("L"))
        self.hashLen = data.eat("L")
        self.hmac = data.eat_length_and_string("L")
        self.cipherText = data.eat_length_and_string("L")

        # For HMAC computation
        self.blob = data.raw[blobStart:data.ofs]
        self.sign = data.eat_length_and_string("L")

    def decrypt(self, masterkey, entropy=None, strongPassword=None, smartCardSecret=None):
        """Try to decrypt the blob. Returns True/False
        :rtype : bool
        :param masterkey: decrypted masterkey value
        :param entropy: optional entropy for decrypting the blob
        :param strongPassword: optional password for decrypting the blob
        :param smartCardSecret: MS Next Gen Crypto secret (e.g. from PIN code)
        """
        for algo in [crypto.CryptSessionKeyType1, crypto.CryptSessionKeyType2]:
            sessionkey = algo(masterkey, self.salt, self.hashAlgo, entropy=entropy, smartcardsecret=smartCardSecret, strongPassword=strongPassword)
            key = crypto.CryptDeriveKey(sessionkey, self.cipherAlgo, self.hashAlgo)
            #RC4 is a stream cipher, and so we need to call module without the mode parameter
            if self.cipherAlgo == "RC4":
              cipher =  self.cipherAlgo.module.new(key[:int(self.cipherAlgo.keyLength)],
                                                  IV=b'\x00' * int(self.cipherAlgo.ivLength))
            else:
              cipher = self.cipherAlgo.module.new(key[:int(self.cipherAlgo.keyLength)],
                                                mode=self.cipherAlgo.module.MODE_CBC,
                                                IV=b'\x00' * int(self.cipherAlgo.ivLength))
            self.cleartext = cipher.decrypt(self.cipherText)
            padding = self.cleartext[-1]
            if padding <= self.cipherAlgo.blockSize:
                self.cleartext = self.cleartext[:-padding]
            # check against provided HMAC
            self.signComputed = algo(masterkey, self.hmac, self.hashAlgo, entropy=entropy, smartcardsecret=smartCardSecret, verifBlob=self.blob)
            self.decrypted = self.signComputed == self.sign
            if self.decrypted:
                return True
        self.decrypted = False
        return self.decrypted

    def __repr__(self):
        s = ["DPAPI BLOB",
             "\n".join(("\tversion      = %(version)d",
                        "\tprovider     = %(provider)s",
                        "\tmkey         = %(mkguid)s",
                        "\tflags        = %(flags)#x",
                        "\tdescr        = %(description)s",
                        "\tcipherAlgo   = %(cipherAlgo)r",
                        "\thashAlgo     = %(hashAlgo)r")) % self.__dict__,
             "\tsalt         = %s" % self.salt.hex(),
             "\thmac         = %s" % self.hmac.hex(),
             "\tcipher       = %s" % self.cipherText.hex(),
             "\tsign         = %s" % self.sign.hex()]
        if self.signComputed is not None:
            s.append("\tsignComputed = %s" % self.signComputed.hex())
        if self.cleartext is not None:
            s.append("\tcleartext    = %r" % self.cleartext)
        return "\n".join(s)

# vim:ts=4:expandtab:sw=4



#!/usr/bin/python3
# -*- coding: utf-8 -*-

#############################################################################
##                                                                         ##
## This file is part of DPAPIck                                            ##
## Windows DPAPI decryption & forensic toolkit                             ##
##                                                                         ##
##                                                                         ##
## Copyright (C) 2010, 2011 Cassidian SAS. All rights reserved.            ##
## Copyright (C) 2023       Insecurity. All rights reserved.               ##
##                                                                         ##
##  Author:  Jean-Michel Picod <jmichel.p@gmail.com>                       ##
##  Updated: Photubias <info@insecurity.be>                                ##
##                                                                         ##
## This program is distributed under GPLv3 licence (see LICENCE.txt)       ##
##                                                                         ##
#############################################################################

from dpapick3 import blob, eater

class DPAPIProbe(eater.DataStruct):
    """This is the generic class for building DPAPIck probes.
        All probes must inherit this class.

    """
    def __init__(self, raw=None):
        """Constructs a DPAPIProbe object.
            If raw is set, automatically builds a DataStruct with that
            and calls parse() method with this.

        """
        self.dpapiblob = None
        self.cleartext = None
        self.entropy = None
        eater.DataStruct.__init__(self, raw)

    def parse(self, data):
        """Parses raw data into structured data.
            Automatically called by __init__. You should not call it manually.

            data is a DataStruct object.

        """
        self.dpapiblob = blob.DPAPIBlob(data.remain())

    def preprocess(self, **k):
        """Optional. Apply tranformations to data before the decryption loop."""
        self.entropy = k.get("entropy")

    def postprocess(self, **k):
        """Optional. Apply transformations after a successful decryption."""
        if self.dpapiblob.decrypted:
            self.cleartext = self.dpapiblob.cleartext

    def try_decrypt_system(self, mkeypool, **k):
        """Decryption loop for SYSTEM account protected blob. eg. wifi blobs.
            Basic probes should not overload this function.

            Returns True/False upon decryption success/failure.

        """
        self.preprocess(**k)
        mkeypool.try_credential(None, None)
        for kguid in self.dpapiblob.guids:
            mks = mkeypool.getMasterKeys(kguid)
            for mk in mks:
                if mk.decrypted:
                    self.dpapiblob.decrypt(mk.get_key(), self.entropy, k.get("strong", None))
                    if self.dpapiblob.decrypted:
                        self.postprocess(**k)
                        return True
        return False

    def try_decrypt_with_hash(self, h, mkeypool, sid, **k):
        """Decryption loop for general blobs with given user's password hash.
            This function will call preprocess() first, then tries to decrypt.

            k may contain optional values such as:
                entropy: the optional entropy to use with that blob.
                strong: strong password given by the user

            Basic probes should not override this one as it contains the full
            decryption logic.

            Returns True/False upon decryption success/failure.

        """
        self.preprocess(**k)
        mkeypool.try_credential_hash(sid, h)
        mks = mkeypool.getMasterKeys(self.dpapiblob.mkguid.encode())
        for mk in mks:
            if mk.decrypted:
                self.dpapiblob.decrypt(mk.get_key(), self.entropy, k.get('strong', None))
                if self.dpapiblob.decrypted:
                    self.postprocess(**k)
                    return True
        return False

    def try_decrypt_with_key(self, key, **k):
        """Decryption loop for general blobs with given decrypted key.
            This function will call preprocess() first, then tries to decrypt.

            k may contain optional values such as:
                entropy: the optional entropy to use with that blob.
                strong: strong password given by the user

            Basic probes should not override this one as it contains the full
            decryption logic.

            Returns True/False upon decryption success/failure.

        """
        self.preprocess(**k)
        
        self.dpapiblob.decrypt(key, self.entropy, k.get('strong', None))
        if self.dpapiblob.decrypted:
            self.postprocess(**k)
            return True
        return False

    def try_decrypt_with_password(self, password, mkeypool, sid, **k):
        """Decryption loop for general blobs with given user's password.
            Simply computes the hash then calls try_decrypt_with_hash()

            Return True/False upon decryption success/failure.

        """
        self.preprocess(**k)
        mkeypool.try_credential(sid, password)

        mks = mkeypool.getMasterKeys(bytes(self.dpapiblob.mkguid,"utf8"))
        for mk in mks:
            if mk.decrypted:
                self.dpapiblob.decrypt(mk.get_key(), self.entropy, k.get("strong", None))
                if self.dpapiblob.decrypted:
                    self.postprocess(**k)
                    return True
        return False

    def __repr__(self):
        s = ["DPAPI default blob", "    %r" % self.dpapiblob]
        return "\n".join(s)

# vim:ts=4:expandtab:sw=4



#!/usr/bin/python3
# -*- coding: utf-8 -*-

#############################################################################
##                                                                         ##
## This file is part of DPAPIck                                            ##
## Windows DPAPI decryption & forensic toolkit                             ##
##                                                                         ##
##                                                                         ##
## Copyright (C) 2010, 2011 Cassidian SAS. All rights reserved.            ##
## Copyright (C) 2023       Insecurity. All rights reserved.               ##
##                                                                         ##
##  Author:  Jean-Michel Picod <jmichel.p@gmail.com>                       ##
##  Updated: Photubias <info@insecurity.be>                                ##
##                                                                         ##
## This program is distributed under GPLv3 licence (see LICENCE.txt)       ##
##                                                                         ##
#############################################################################

import array, hashlib, hmac, struct
from Crypto.Cipher import AES, ARC4, DES, DES3

class CryptoAlgo(object):
    """This class is used to wrap Microsoft algorithm IDs with M2Crypto"""

    class Algo(object):
        def __init__(self, data):
            self.data = data

        def __getattr__(self, attr):
            if attr in self.data:
                return self.data[attr]
            raise AttributeError(attr)

    _crypto_data = {}

    @classmethod
    def add_algo(cls, algnum, **kargs):
        cls._crypto_data[algnum] = cls.Algo(kargs)
        if 'name' in kargs:
            kargs['ID'] = algnum
            cls._crypto_data[kargs['name']] = cls.Algo(kargs)

    @classmethod
    def get_algo(cls, algnum):
        return cls._crypto_data[algnum]

    def __init__(self, i):
        self.algnum = i
        self.algo = CryptoAlgo.get_algo(i)

    name = property(lambda self: self.algo.name)
    module = property(lambda self: self.algo.module)
    keyLength = property(lambda self: self.algo.keyLength / 8)
    ivLength = property(lambda self: self.algo.IVLength / 8)
    blockSize = property(lambda self: self.algo.blockLength / 8)
    digestLength = property(lambda self: self.algo.digestLength / 8)

    def do_fixup_key(self, key):
        try:
            return self.algo.keyFixup.__call__(key)
        except AttributeError:
            return key

    def __repr__(self):
        return "%s [%#x]" % (self.algo.name, self.algnum)


def des_set_odd_parity(key):
    _lut = [1, 1, 2, 2, 4, 4, 7, 7, 8, 8, 11, 11, 13, 13, 14, 14, 16, 16, 19,
            19, 21, 21, 22, 22, 25, 25, 26, 26, 28, 28, 31, 31, 32, 32, 35, 35, 37,
            37, 38, 38, 41, 41, 42, 42, 44, 44, 47, 47, 49, 49, 50, 50, 52, 52, 55,
            55, 56, 56, 59, 59, 61, 61, 62, 62, 64, 64, 67, 67, 69, 69, 70, 70, 73,
            73, 74, 74, 76, 76, 79, 79, 81, 81, 82, 82, 84, 84, 87, 87, 88, 88, 91,
            91, 93, 93, 94, 94, 97, 97, 98, 98, 100, 100, 103, 103, 104, 104, 107,
            107, 109, 109, 110, 110, 112, 112, 115, 115, 117, 117, 118, 118, 121,
            121, 122, 122, 124, 124, 127, 127, 128, 128, 131, 131, 133, 133, 134,
            134, 137, 137, 138, 138, 140, 140, 143, 143, 145, 145, 146, 146, 148,
            148, 151, 151, 152, 152, 155, 155, 157, 157, 158, 158, 161, 161, 162,
            162, 164, 164, 167, 167, 168, 168, 171, 171, 173, 173, 174, 174, 176,
            176, 179, 179, 181, 181, 182, 182, 185, 185, 186, 186, 188, 188, 191,
            191, 193, 193, 194, 194, 196, 196, 199, 199, 200, 200, 203, 203, 205,
            205, 206, 206, 208, 208, 211, 211, 213, 213, 214, 214, 217, 217, 218,
            218, 220, 220, 223, 223, 224, 224, 227, 227, 229, 229, 230, 230, 233,
            233, 234, 234, 236, 236, 239, 239, 241, 241, 242, 242, 244, 244, 247,
            247, 248, 248, 251, 251, 253, 253, 254, 254]
    tmp = array.array("B")
    tmp.fromstring(key)
    for i, v in enumerate(tmp):
        tmp[i] = _lut[v]
    return tmp.tostring()


CryptoAlgo.add_algo(0x6603, name="DES3", keyLength=192, IVLength=64, blockLength=64, module=DES3,
                    keyFixup=des_set_odd_parity)
CryptoAlgo.add_algo(0x6611, name="AES", keyLength=128, IVLength=128, blockLength=128, module=AES)
CryptoAlgo.add_algo(0x660e, name="AES-128", keyLength=128, IVLength=128, blockLength=128, module=AES)
CryptoAlgo.add_algo(0x660f, name="AES-192", keyLength=192, IVLength=128, blockLength=128, module=AES)
CryptoAlgo.add_algo(0x6610, name="AES-256", keyLength=256, IVLength=128, blockLength=128, module=AES)
CryptoAlgo.add_algo(0x6601, name="DES", keyLength=64, IVLength=64, blockLength=64, module=DES,
                    keyFixup=des_set_odd_parity)
CryptoAlgo.add_algo(0x6801, name="RC4", keyLength=40, IVLength=128, blockLength=1, module=ARC4)

CryptoAlgo.add_algo(0x8009, name="HMAC", digestLength=160, blockLength=512)

CryptoAlgo.add_algo(0x8003, name="md5", digestLength=128, blockLength=512)

CryptoAlgo.add_algo(0x8004, name="sha1", digestLength=160, blockLength=512)
CryptoAlgo.add_algo(0x800c, name="sha256", digestLength=256, blockLength=512)
CryptoAlgo.add_algo(0x800d, name="sha384", digestLength=384, blockLength=1024)
CryptoAlgo.add_algo(0x800e, name="sha512", digestLength=512, blockLength=1024)


def CryptSessionKeyType1(masterkey, nonce, hashAlgo, entropy=None, strongPassword=None, smartcardsecret=None, verifBlob=None):
    """Computes the decryption key for Type1 DPAPI blob, given the masterkey and optional information.

    This implementation relies on a faulty implementation from Microsoft that does not respect the HMAC RFC.
    Instead of updating the inner pad, we update the outer pad...
    This algorithm is also used when checking the HMAC for integrity after decryption

    :param masterkey: decrypted masterkey (should be 64 bytes long)
    :param nonce: this is the nonce contained in the blob or the HMAC in the blob (integrity check)
    :param entropy: this is the optional entropy from CryptProtectData() API
    :param strongPassword: optional password used for decryption or the blob itself
    :param smartcardsecret: optional MS Next Gen Crypto secret (e.g. from PIN code)
    :param verifBlob: optional encrypted blob used for integrity check
    :returns: decryption key
    :rtype : str
    """
    if len(masterkey) > 20: masterkey = hashlib.sha1(masterkey).digest()

    masterkey += ('\x00' * int(hashAlgo.blockSize)).encode()
    pad1 = ''.join(chr(masterkey[i] ^ 0x36) for i in range(int(hashAlgo.blockSize)))
    pad2 = ''.join(chr(masterkey[i] ^ 0x5c) for i in range(int(hashAlgo.blockSize)))

    digest1 = hashlib.new(hashAlgo.name)
    digest1.update(pad2.encode('latin1'))
    
    digest2 = hashlib.new(hashAlgo.name)
    digest2.update(pad1.encode('latin1'))
    digest2.update(nonce)
    if smartcardsecret is not None:
         digest2.update(entropy + smartcardsecret)
         if verifBlob is not None: digest2.update(verifBlob)
    
    digest1.update(digest2.digest())
    if entropy is not None and smartcardsecret is None: digest1.update(entropy)
    if strongPassword is not None:
        strongPassword = hashlib.sha1(strongPassword.rstrip("\x00").encode("UTF-16LE")).digest()
        digest1.update(strongPassword)
    if smartcardsecret is None and verifBlob is not None: digest1.update(verifBlob)
    
    return digest1.digest()


def CryptSessionKeyType2(masterkey, nonce, hashAlgo, entropy=None, strongPassword=None, smartcardsecret=None, verifBlob=None):
    """Computes the decryption key for Type2 DPAPI blob, given the masterkey and optional information.

    This implementation relies on an RFC compliant HMAC implementation
    This algorithm is also used when checking the HMAC for integrity after decryption

    :param masterkey: decrypted masterkey (should be 64 bytes long)
    :param nonce: this is the nonce contained in the blob or the HMAC in the blob (integrity check)
    :param entropy: this is the optional entropy from CryptProtectData() API
    :param strongPassword: optional password used for decryption or the blob itself
    :param smartcardsecret: optional MS Next Gen Crypto secret (e.g. from PIN code)
    :param verifBlob: optional encrypted blob used for integrity check
    :returns: decryption key
    :rtype : str
    """
    if len(masterkey) > 20: masterkey = hashlib.sha1(masterkey).digest()

    digest = hmac.new(masterkey, digestmod=getDigestmod(hashAlgo.name))
    digest.update(nonce)
    if entropy is not None:
        digest.update(entropy)
    if strongPassword is not None:
        strongPassword = hashlib.sha512(strongPassword.rstrip("\x00").encode("UTF-16LE")).digest()
        digest.update(strongPassword)
    elif verifBlob is not None:
        digest.update(verifBlob)
    return digest.digest()


def CryptDeriveKey(h, cipherAlgo, hashAlgo):
    """Internal use. Mimics the corresponding native Microsoft function"""
    if len(h) > hashAlgo.blockSize:
        h = hashlib.new(hashAlgo.name, h).digest()
    if len(h) >= cipherAlgo.keyLength:
        return h
    h += ('\x00' * int(hashAlgo.blockSize)).encode()
    pad1 = ''.join(chr(h[i] ^ 0x36) for i in range(int(hashAlgo.blockSize)))
    pad2 = ''.join(chr(h[i] ^ 0x5c) for i in range(int(hashAlgo.blockSize)))
    k = hashlib.new(hashAlgo.name, pad1.encode('latin1')).digest() + hashlib.new(hashAlgo.name, pad2.encode('latin1')).digest()
    k = cipherAlgo.do_fixup_key(k)
    return k


def decrypt_lsa_key_nt5(lsakey, syskey):
    """This function decrypts the LSA key using the syskey"""
    dg = hashlib.md5()
    dg.update(syskey)
    for i in range(1000):
        dg.update(lsakey[60:76])
    arcfour = ARC4.new(dg.digest())
    deskey = arcfour.decrypt(lsakey[12:60])
    return [deskey[16 * x:16 * (x + 1)] for x in range(3)]


def decrypt_lsa_key_nt6(lsakey, syskey):
    """This function decrypts the LSA keys using the syskey"""
    dg = hashlib.sha256()
    dg.update(syskey)
    for i in range(1000):
        dg.update(lsakey[28:60])
    keys = AES.new(dg.digest(),AES.MODE_ECB).decrypt(lsakey[60:])
    size = struct.unpack_from("<L", keys)[0]
    keys = keys[16:16 + size]
    currentkey = "%0x-%0x-%0x-%0x%0x-%0x%0x%0x%0x%0x%0x" % struct.unpack("<L2H8B", keys[4:20])
    nb = struct.unpack("<L", keys[24:28])[0]
    off = 28
    kd = {}
    for i in range(nb):
        g = "%0x-%0x-%0x-%0x%0x-%0x%0x%0x%0x%0x%0x" % struct.unpack("<L2H8B", keys[off:off + 16])
        t, l = struct.unpack_from("<2L", keys[off + 16:])
        k = keys[off + 24:off + 24 + l]
        kd[g] = {"type": t, "key": k}
        off += 24 + l
    return (currentkey, kd)


def SystemFunction005(secret, key):
    """This function is used to decrypt LSA secrets.
    Reproduces the corresponding Windows internal function.
    Taken from creddump project https://code.google.com/p/creddump/
    """
    decrypted_data = ''
    j = 0
    algo = CryptoAlgo(0x6603)
    for i in range(0, len(secret), 8):
        enc_block = secret[i:i + 8]
        block_key = key[j:j + 7]
        des_key = []
        des_key.append(ord(block_key[0]) >> 1)
        des_key.append(((ord(block_key[0]) & 0x01) << 6) | (ord(block_key[1]) >> 2))
        des_key.append(((ord(block_key[1]) & 0x03) << 5) | (ord(block_key[2]) >> 3))
        des_key.append(((ord(block_key[2]) & 0x07) << 4) | (ord(block_key[3]) >> 4))
        des_key.append(((ord(block_key[3]) & 0x0F) << 3) | (ord(block_key[4]) >> 5))
        des_key.append(((ord(block_key[4]) & 0x1F) << 2) | (ord(block_key[5]) >> 6))
        des_key.append(((ord(block_key[5]) & 0x3F) << 1) | (ord(block_key[6]) >> 7))
        des_key.append(ord(block_key[6]) & 0x7F)
        des_key = algo.do_fixup_key("".join([chr(x << 1) for x in des_key]))

        decrypted_data += DES.new(des_key).decrypt(enc_block)
        j += 7
        if len(key[j:j + 7]) < 7:
            j = len(key[j:j + 7])
    dec_data_len = struct.unpack("<L", decrypted_data[:4])[0]
    return decrypted_data[8:8 + dec_data_len]


def decrypt_lsa_secret(secret, lsa_keys):
    """This function replaces SystemFunction005 for newer Windows"""
    keyid = "%0x-%0x-%0x-%0x%0x-%0x%0x%0x%0x%0x%0x" % struct.unpack("<L2H8B", secret[4:20])
    if keyid not in lsa_keys:
        return None
    algo = struct.unpack("<L", secret[20:24])[0]
    dg = hashlib.sha256()
    dg.update(lsa_keys[keyid]['key'])
    for i in range(1000):
        dg.update(secret[28:60])
    clear = AES.new(dg.digest(),AES.MODE_ECB).decrypt(secret[60:])
    size = struct.unpack_from("<L", clear)[0]
    return clear[16:16 + size]


def pbkdf2(passphrase, salt, keylen, iterations, digest='sha1'):
    """Implementation of PBKDF2 that allows specifying digest algorithm.
    Returns the corresponding expanded key which is keylen long.
    """
    buff = b''
    i = 1
    while len(buff) < keylen:
        U = salt + struct.pack("!L", i)
        i += 1
        derived = hmac.new(passphrase, U, digestmod=getDigestmod(digest)).digest()
        for r in range(iterations - 1):
            actual = hmac.new(passphrase, derived, digestmod=getDigestmod(digest)).digest()
            derived = ''.join([chr(int(x,16) ^ int(y,16)) for (x, y) in zip(derived.hex(), actual.hex())]).encode().hex()
            result = ''
            for j in range(len(derived)): 
                if j % 2 == 1: result += derived[j]
            derived = bytes.fromhex(result)
        buff += derived
    return buff[:keylen]

def getDigestmod(digest):
    Dmod = {
        'sha1':hashlib.sha1,
        'sha224':hashlib.sha224,
        'sha256':hashlib.sha256,
        'sha384':hashlib.sha384,
        'sha512':hashlib.sha512
    }
    return Dmod.get(digest,hashlib.sha1)

def derivePwdHash(pwdhash, userSID, digest='sha1'):
    """Internal use. Computes the encryption key from a user's password hash"""
    #return hmac.new(pwdhash, (userSID + "\0").encode("UTF-16LE"), digestmod=lambda: hashlib.new(digest)).digest()
    return hmac.new(pwdhash, (userSID + "\0").encode("UTF-16LE"), digestmod=getDigestmod(digest)).digest()


def dataDecrypt(cipherAlgo, hashAlgo, raw, encKey, iv, rounds):
    """Internal use. Decrypts data stored in DPAPI structures."""
    hname = {"HMAC": "sha1"}.get(hashAlgo.name, hashAlgo.name)
    derived = pbkdf2(encKey, iv, int(cipherAlgo.keyLength + cipherAlgo.ivLength), rounds, hname)
    key, iv = derived[:int(cipherAlgo.keyLength)], derived[int(cipherAlgo.keyLength):]
    key = key[:int(cipherAlgo.keyLength)]
    iv = iv[:int(cipherAlgo.ivLength)]
    cipher = cipherAlgo.module.new(key, mode=cipherAlgo.module.MODE_CBC, IV=iv)
    cleartxt = cipher.decrypt(raw)
    return cleartxt


def DPAPIHmac(hashAlgo, pwdhash, hmacSalt, value):
    """Internal function used to compute HMACs of DPAPI structures"""
    hname = {"HMAC": "sha1"}.get(hashAlgo.name, hashAlgo.name)
    #encKey = hmac.new(pwdhash, digestmod=lambda: hashlib.new(hname))
    encKey = hmac.new(pwdhash, digestmod=getDigestmod(hname))
    encKey.update(hmacSalt)
    encKey = encKey.digest()
    #rv = hmac.new(encKey, digestmod=lambda: hashlib.new(hname))
    rv = hmac.new(encKey, digestmod=getDigestmod(hname))
    rv.update(value)
    return rv.digest()

# vim:ts=4:expandtab:sw=4



#!/usr/bin/python3
# -*- coding: utf-8 -*-

#############################################################################
##                                                                         ##
## This file is part of DPAPIck                                            ##
## Windows DPAPI decryption & forensic toolkit                             ##
##                                                                         ##
##                                                                         ##
## Copyright (C) 2010, 2011 Cassidian SAS. All rights reserved.            ##
## Copyright (C) 2023, 2024 Insecurity. All rights reserved.               ##
##                                                                         ##
##  Author:  Jean-Michel Picod <jmichel.p@gmail.com>                       ##
##  Updated: Photubias <info@insecurity.be>                                ##
##                                                                         ##
## This program is distributed under GPLv3 licence (see LICENCE.txt)       ##
##                                                                         ##
#############################################################################

import setuptools

long_description = open('README.md', 'r').read()

setuptools.setup(
    name = 'dpapick3',
    version = '0.5.1',
    author = 'Tijl Deneut',
    author_email = 'info@insecurity.be',
    description = 'A native implementation of DPAPI',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/tijldeneut/dpapick3',
    packages = setuptools.find_packages(),
    include_package_data = True,
    zip_safe = False,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires = '>=3.2',
    install_requires = [
        'pycryptodome',
        'python-registry',
        'pyasn1'
    ]
)
